# -*- coding: utf-8 -*-

"""cultaptation engine  version 0.3

This program is an implementation of the cultaptation model, a model
for studying the evolution of social learning by Kevin Laland and
Luke Rendel (University of St Andrews), as described in:
http://www.intercult.su.se/cultaptation/tournament_details.php

Copyright (c) 2008 by Eckhart Arnold (www.eckhartarnold.de)
                   with strategies from Julia Lohmann, Philip Schächtele,
                   Christoph Siemroth and Martin Neumann
                   
License: GNU Public License V3
(http://www.gnu.de/documents/gpl-3.0.en.html)

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the 
Free Software Foundation, either version 3 of the License, or (at your 
option) any later version.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along 
with this program. If not, see <http://www.gnu.org/licenses/>. 
"""


import os
import Tkinter as Tk
import tkMessageBox
import Engine, Strategy, Monitor, Configuration

try: from PyPlotter import tkGfx
except ImportError: pass

# psyco support to speed up the program
# (delete or comment out the following lines,
#  if it takes too much memory!)
#try:
#    import psyco
#    psyco.profile()
#    print "Using psyco JIT compiler to speed up program..."
#except ImportError:
#    print "Psyco JIT not present. Will be slower..."

def ScanStrategies():
    strategyList = []
    for objStr in dir(Strategy):
        if objStr == "Stuffing": continue
        obj = eval("Strategy."+objStr)
        if type(obj) == type(Strategy.ExampleStrategy) and \
           obj.func_code.co_varnames[0:3] == \
           Strategy.ExampleStrategy.func_code.co_varnames[0:3]:
            strategyList.append(obj)
    return strategyList

class ApplicationWindow(Tk.Tk):
    def __init__(self):
        Tk.Tk.__init__(self)
        self.title("cultaptation engine (Python version)")
        self.protocol("WM_DELETE_WINDOW", self.onClose)
        
        self.ctrl = Tk.Frame(self, bd=4)
        self.ctrl.pack(side = Tk.BOTTOM, fill = Tk.X)        
        
        self.frow = Tk.Frame(self, bd = 4)
        self.frow.pack(side = Tk.BOTTOM) 
        
        self.stgy = Tk.Frame(self)
        self.stgy.pack(side = Tk.LEFT, fill = Tk.Y)
        
        self.help = Tk.Text(self, background="Gold")
        self.help.pack(side = Tk.RIGHT, fill = Tk.BOTH)
        
        self.flags = {"Live Plot": Tk.BooleanVar(value = Configuration.LIVE_PLOT), 
                      "Frequency Graph": Tk.BooleanVar(value = Configuration.REGISTER_FREQUENCIES), 
                      "Post Mortem Analysis": Tk.BooleanVar(value = Configuration.POST_MORTEM_ANALYSIS != 0)}
        Tk.Label(self.frow, text="Flags:").pack(side=Tk.LEFT)       
        for f in self.flags.keys():
            Tk.Checkbutton(self.frow, text = f, var = self.flags[f]).pack(side=Tk.LEFT)
        
        self.strategyList = ScanStrategies()
        self.rack = {}
        Tk.Label(self.stgy, text="Strategies:").pack()
        for s in self.strategyList:
            self.rack[s] = Tk.BooleanVar()
            button = Tk.Checkbutton(self.stgy, text=s.func_name, anchor=Tk.W,
                                    var = self.rack[s])
            button.select()
            button.pack(expand = Tk.YES, fill = Tk.BOTH)
            button.bind("<Enter>", lambda ev, st = s: self.onEnter(st))
            button.bind("<Leave>", lambda st = s: self.onLeave(st))
        
        Tk.Button(self.stgy, text="Select All", command = self.onSelectAll).\
            pack(side=Tk.LEFT)
        Tk.Button(self.stgy, text="Select None", command = self.onSelectNone).\
            pack(side=Tk.LEFT)

        Tk.Button(self.ctrl, text="-> Run Match", command = self.onRun).\
            pack(side = Tk.LEFT)
        Tk.Button(self.ctrl, text="-> Run Contest", 
                  command = self.onContest).pack(side = Tk.LEFT)
        Tk.Button(self.ctrl, text="-> Run Tournament", 
                  command = self.onSimulation).pack(side = Tk.LEFT)
        Tk.Label(self.ctrl, text=" ").pack(side = Tk.LEFT)
        self.progressBar = Tk.Canvas(self.ctrl, bd = 2, relief = Tk.SUNKEN, 
                                     height = 16)
        self.progressBar.pack(side = Tk.LEFT, expand = Tk.YES, fill = Tk.X)
        
        self.liveWindow = None
        self.canvas = None
        self.selectedStrategies = []
        self.simRunning = False
        self.maxMatches = 0
        self.rememberProgress = -1
        Monitor.progressCallback = self.onProgress
        
        self.mainloop()
        
    def showHTMLPage(self, path):
        try:
            os.startfile(path)
        except AttributeError:
            os.system("firefox "+path+" &")
        
    def onEnter(self, strategy):
        self.help.delete('1.0', Tk.END)        
        self.help.insert('1.0', strategy.func_name+"\n\n"+strategy.__doc__)
        
    def onLeave(self, strategy):
        self.help.delete('1.0', Tk.END)
        self.help.insert('1.0', __doc__)
        
    def onSelectAll(self):
        for s in self.rack.keys():
            self.rack[s].set(True)
    
    def onSelectNone(self):
        for s in self.rack.keys():
            self.rack[s].set(False)        
        
    def prepareSim(self):
        """Checks whether a simulation is not already running and whether at
        least one strategy has been selected (displays a warning message and 
        returns False if any of these conditions is false), collects the 
        selected strategies in the list 'self.selectedStrategies', opens a 
        window for the live plot, if possible and not already done.
        Sets the 'simRunning' flag and returns True, if everything is fine."""
        if self.simRunning:
            tkMessageBox.showwarning("Simulation already running!",
                "A simulation is already running. If you do not want to "+\
                "wait until the simulation is finished then close and "+\
                "restart the application.")
            return False
        self.selectedStrategies = [s for s in self.rack.keys() \
                                   if self.rack[s].get()] 
        if len(self.selectedStrategies) < 1:
            tkMessageBox.showwarning("No strategies selected!?", 
                        "Please select one or more strategies to proceed")                
            return False
        if self.flags["Live Plot"].get(): Configuration.LIVE_PLOT = 1
        else: Configuration.LIVE_PLOT = 0
        if self.flags["Frequency Graph"].get(): Configuration.FREQUENCY_GRAPH = 1
        else: Configuration.REGISTER_FREQUENCIES = 0
        if self.flags["Post Mortem Analysis"].get(): 
            Configuration.POST_MORTEM_ANALYSIS = 1
        else: Configuration.POST_MORTEM_ANALYSIS = 0
        if Configuration.LIVE_PLOT and Monitor.HAS_PYPLOTTER:
            if not self.liveWindow:
                self.liveWindow = Tk.Toplevel(width=800, height=460)
                self.liveWindow.title("Live Plot")
                self.canvas = Tk.Canvas(self.liveWindow, width=800, height=460)
                self.canvas.pack()
                self.liveWindow.update()
                Monitor.gfx = tkGfx.Driver(self.canvas)
                Monitor.gfx.refresh = lambda : self.liveWindow.update()
                Monitor.rlgCallback = lambda : self.canvas.delete("all")
            else:
                Monitor.resetLiveGraph()    
        self.simRunning = True
        return True
   
    def finishSim(self):
        """Cleans up after a simulation is finished and resets the 'simRunning'
        flag."""
        if self.canvas:
            self.canvas.destroy()
            self.canvas = None
        if self.liveWindow:
            self.liveWindow.destroy()
            self.liveWindow = None
        self.simRunning = False
            
    def onRun(self):
        if not self.prepareSim(): return
        if len(self.selectedStrategies) == 2:
            self.maxMatches = 2
            result = Engine.PairwiseMatch(self.selectedStrategies[0],
                                          self.selectedStrategies[1])
            Monitor.CreateGenericDir()            
            path = Monitor.PCRecord(result)
            self.showHTMLPage(path)              
            Monitor.resetLiveGraph()
            result = Engine.PairwiseMatch(self.selectedStrategies[1],
                                          self.selectedStrategies[0])
            path = Monitor.PCRecord(result)
            self.showHTMLPage(path) 
        else:
            self.maxMatches = 1
            result = Engine.MeleeMatch(self.selectedStrategies)
            Monitor.CreateGenericDir()
            path = Monitor.MCRecord(result)
            self.showHTMLPage(path)                 
        self.finishSim()       
        
    def onContest(self):
        if not self.prepareSim(): return
        Monitor.CreateGenericDir()
        if len(self.selectedStrategies) == 2:
            self.maxMatches = Configuration.PW_REPETITIONS
            result = Engine.PairwiseContest(self.selectedStrategies[0],
                                            self.selectedStrategies[1])
            path = Monitor.PCFinished(result)
        else:
            self.maxMatches = Configuration.MELEE_REPETITIONS
            result = Engine.MeleeContest(self.selectedStrategies)
            path = Monitor.MCFinished(result)
        self.showHTMLPage(path)
        self.finishSim()
    
    def onSimulation(self):
        if not self.prepareSim(): return
        Monitor.PrepareTournamentDirSet()
        n = len(self.selectedStrategies)
        self.maxMatches = Configuration.MELEE_REPETITIONS + \
            Configuration.PW_REPETITIONS * n * (n-1) / 2
        result = Engine.FullTournament(self.selectedStrategies)
        path = Monitor.TournamentFinished(result)
        self.showHTMLPage(path) 
        self.finishSim()
                
    def onProgress(self, matchNr, roundNr):
        w = int(self.progressBar.cget("width"))
        h = int(self.progressBar.cget("height"))
        progress = ((matchNr-1) * Configuration.MATCH_LENGTH + roundNr) * w / \
                   (self.maxMatches * Configuration.MATCH_LENGTH)
        if progress != self.rememberProgress:
            self.rememberProgress = progress
            self.progressBar.delete("all")            
            self.progressBar.create_rectangle(3, 3, progress, h+2, fill = "blue")
            self.update()
        
    def onClose(self):
        self.destroy()
        self.quit()
    

if __name__ == "__main__":
    app = ApplicationWindow()
    
    